#! /usr/bin/python

from __future__ import division
import numpy as np
import cPickle
import re
import pylab
import itertools
import os
import scipy.io
import scipy.misc
import scipy.ndimage # label_im, nb_labels = scipy.ndimage.label(area)

class NN:
   def __init__(self, def_file, sigma):

      self.sigma = sigma
      # this parameter has minor impact on RMSE
      # because it defines how a patch is looked at by the neural network
      # the data provider in GPU powered training written in python is row-major
      # the data provider in CPU powered training written in Eigen is column-major
      self.transpose = False

      if def_file.endswith(".mat"):
         # parse the German NNs (two versions)
         connections = scipy.io.loadmat(def_file)
         try:
            # the linear output is not considered hidden
            self.nHiddenLayers = connections["w"].shape[1] - 1
            self.connections = [connections["w"][0][k] for k in range(self.nHiddenLayers+1)]
         except:
            self.nHiddenLayers = len([key for key in connections.keys() if "w_" in key]) - 1
            self.connections = [connections["w_"+str(k+1)] for k in range(self.nHiddenLayers+1)]
         # input and output atoms are of interest
         self.iSize = int(np.sqrt(self.connections[0].shape[1] - 1))
         self.oSize = int(np.sqrt(self.connections[-1].shape[0]))

      elif def_file.endswith(".cpp"):
         # parse my cxx implementation
         for statement in re.split(";", open(def_file, "r").read()):
            if "nHiddenLayers" in statement:
               self.nHiddenLayers = int(re.split("=", statement)[1])
            if "patchSize" in statement:
               self.iSize = int(re.split("=", statement)[1])
               # assume this condition
               self.oSize = self.iSize
            if "optStackTheta" in statement:
               s = re.search(r"={(.*)}", statement.replace("\n", "")).group(1)
               theta = [float(x) for x in s.split(",")]
            if "hiddenSize" in statement:
               # assume a constant number of hidden nodes in intermediary layers
               hiddenSize = int(re.split("=", statement)[1])
         self.connections = [];
         # output layer
         odSize = self.oSize**2
         end = odSize*hiddenSize;
         W = np.asarray(theta[:end]).reshape(hiddenSize, odSize).T
         b = np.asarray(theta[end:end+odSize]).reshape(odSize, 1)
         self.connections.append(np.append(W, b, 1))
         # hidden layer
         unit = (hiddenSize+1)*hiddenSize;
         for l in range(self.nHiddenLayers, 1, -1):
            start = (hiddenSize+1)*odSize + (self.nHiddenLayers-l)*unit
            W = np.asarray(theta[start:start+hiddenSize*hiddenSize]).reshape(hiddenSize, hiddenSize).T
            b = np.asarray(theta[start+hiddenSize*hiddenSize:start+hiddenSize*hiddenSize+hiddenSize]).reshape(hiddenSize,1)
            self.connections.append(np.append(W, b, 1))
         # input layer
         idSize = self.iSize**2
         start = (hiddenSize+1)*odSize + (self.nHiddenLayers-1)*unit
         W = np.asarray(theta[start:start+hiddenSize*idSize]).reshape(idSize, hiddenSize).T
         b = np.asarray(theta[start+hiddenSize*idSize:]).reshape(hiddenSize, 1)
         self.connections.append(np.append(W, b, 1))
         self.connections.reverse()
         self.transpose = True

      else:
         # convnet trained NN
         with open(def_file, "rb") as f:
            d = cPickle.load(f)
         self.iSize = d["iSize"]
         self.oSize = d["oSize"]
         self.nHiddenLayers = d["nHiddenLayers"]
         self.connections = [m.astype(np.single) for m in d["connections"]]
         self.transpose = False

      # print
      params = [self.nHiddenLayers+1, self.iSize, self.oSize]
      statement = "\nRead in the {0}-layer NN whose input and output patch size is {1}x{1} and {2}x{2}"
      print(statement.format(*params))
      params = [self.connections[k].shape[0] for k in range(self.nHiddenLayers)]
      params.append(self.connections[-1].shape[1] - 1)
      statement = "There are " + "{}, "*self.nHiddenLayers + "{} nodes at each layer leading to the output\n"
      print(statement.format(*params))
      self.name = "NN_{}_{}_{}".format(self.iSize, self.oSize, self.nHiddenLayers)

   def showFeatures(self, which):
      # show the atoms from either the input or the output layer
      if which == "in":
         total = self.connections[0].shape[0]
         atoms = self.connections[0][:total, :-1].T
      else:
         total = self.connections[-1].shape[1]-1
         atoms = self.connections[-1][:, :total]
      title = "learned features at the {} layer".format("input" if which == "in" else "output")
      NN.showAtoms(atoms, title)

   def showActivations(self, type="cdf"):
      # the distribution of a typical input's activations
      clean = np.ones((self.iSize**2, 1))*127
      activations = self.feedforward(clean, self.sigma, predict=False)
      if type == "cdf":
         NN.showCDF(activations, "output activations")
      else:
         pylab.hist(activations, bins=1e2)
         pylab.title("histogram of output activations")
         pylab.show()

   def showNN(self, type="cdf"):
      # one liner for a first look at the neural network
      nn.showFeatures("in")
      nn.showFeatures("out")
      nn.showActivations(type=type)

   def feedforward(self, patches, sigma, areClean=True, optMean=None, predict=True):
      ratio = self.sigma/sigma
      parts = patches + np.random.normal(scale=sigma, size=patches.shape) if areClean else patches
      mode = "patch mode" if areClean else "image mode"
      if sigma != self.sigma or optMean is not None:
         useTrick = True
         print("{} run at noise level = {} and optMean = {}".format(mode, sigma, optMean))
         # noise normalization
         parts *= ratio
         # German style input normalization
         parts = (parts/255 - np.ones(parts.shape)*0.5)*5
         # SPL patch mean shift
         shifts = np.mean(parts, axis=0) - optMean
         parts -= np.tile(shifts, [parts.shape[0], 1])
      else:
         useTrick = False
         print("{} run at noise level = {}".format(mode, sigma))
         parts = (parts/255 - np.ones(parts.shape)*0.5)*5
      # neural feedforward for real
      ones = np.ones([1, patches.shape[1]])
      for layer in range(self.nHiddenLayers):
         parts = np.tanh(np.dot(self.connections[layer], np.append(parts, ones, 0)).astype(np.single))
      # shall we stop at the last coding layer?
      if predict:
         parts = np.dot(self.connections[-1], np.append(parts, ones, 0))
         if useTrick:
            parts += np.tile(shifts, [parts.shape[0], 1])
         parts = (parts*0.2 + np.ones(parts.shape)*0.5)*(255/ratio)

      return parts

   def neuralDenoising(self, clean, sigma, name, optMean=None):
      noisy = clean + np.random.normal(scale=sigma, size=clean.shape)
      offset = (self.iSize - self.oSize)/2
      if offset > 0:
         noisy = NN.extendIm(noisy, offset)
      scipy.misc.imsave("noisy.png", noisy)

      patches = self.im2patches(noisy, self.iSize)
      optMean = -0.5 if sigma != self.sigma and optMean is None else optMean
      denoised = self.feedforward(patches, sigma, areClean=False, optMean=optMean)
      denoisedIm = self.patches2im(clean, denoised)
      np.clip(denoisedIm, 0, 255, out=denoisedIm)
      scipy.misc.imsave("pNN_{}_{}".format(sigma, name), denoisedIm.astype("uint8"))

      RMSE = np.linalg.norm((denoisedIm - clean).flatten())/np.sqrt(clean.size)
      print("basic neural denoising results in {} RMSE = {}".format(name.split(".")[0], RMSE))
      return RMSE

   def denoiseEmAll(self, clean_folder_name, sigma, denoised_folder_name="df", freshRun=True, algo_list=["bm3d", "nlbayes"]):
      curr_dir = os.path.dirname(os.path.abspath(__file__))
      denoised_folder = os.path.join(curr_dir, denoised_folder_name)
      clean_folder = os.path.join(curr_dir, clean_folder_name)
      names = sorted(os.listdir(clean_folder))
      if freshRun:
         rmses = ["pNN"]
         for name in names:
            clean = scipy.misc.imread(os.path.join(clean_folder, name)).astype(np.single)
            dname = "pNN_{}_{}".format(sigma, name)
            rmses.append("{0:.2f}".format(self.neuralDenoising(clean, sigma, name)))
            # mv them to the default folder
            os.rename(os.path.join(curr_dir, dname), os.path.join(denoised_folder, dname))
      # print
      header = [name.split(".")[0] for name in names]
      header.insert(0, "algorithm")
      template = "| ".join(["{}{}:{}{}".format("{", k, len(head)+1, "}") for k, head in enumerate(header)])
      print("="*(len(header)*3+sum([len(head) for head in header])))
      print(template.format(*tuple(header)))
      print("-"*(len(header)*3+sum([len(head) for head in header])))
      if freshRun:
         print(template.format(*tuple(rmses)))
      else:
         algo_list.insert(0, "pNN")
      for algo in algo_list:
         rmses = []
         rmses.append(algo)
         for name in names:
            clean = scipy.misc.imread(os.path.join(clean_folder, name)).astype(np.single)
            dname = "{}_{}_{}".format(algo, sigma, name)
            denoised = scipy.misc.imread(os.path.join(denoised_folder, dname)).astype(np.single)
            rmses.append("{0:.2f}".format(np.linalg.norm((denoised - clean).flatten())/np.sqrt(clean.size)))
         print(template.format(*tuple(rmses)))
      print("="*(len(header)*3+sum([len(head) for head in header]))+"\n")

   def patches2im(self, img, patches, normalize=True):
      output = np.zeros(img.shape)
      pSize = int(np.sqrt(patches.shape[0]))
      blocks = patches.T.reshape(patches.shape[1], pSize, pSize)
      rct = [range(img.shape[0]-pSize+1), range(img.shape[1]-pSize+1)]
      for k, rc in enumerate(itertools.product(*rct)):
         r, c = rc
         if self.transpose:
            output[r:r+pSize, c:c+pSize] += blocks[k, ...].T
         else:
            output[r:r+pSize, c:c+pSize] += blocks[k, ...]

      if normalize:
         unit = np.ones([pSize, pSize])
         weight = np.zeros(img.shape)
         for rc in itertools.product(*rct):
            r, c = rc
            weight[r:r+pSize, c:c+pSize] += unit
         return output/weight
      else:
         return output

   def im2patches(self, img, pSize, keepPos=False):
      # memory strides requires this
      img = np.ascontiguousarray(img)
      # make a four dimensional matrix out of img
      # the first two dimension indexes individual patch"s coordinates
      # the last two indexes the pixels within these patches
      shape = (img.shape[0]-pSize+1, img.shape[1]-pSize+1, pSize, pSize)
      # img is inherently a continuum of memory
      # tell where each pixel lies in the previouly defined structure
      # python is row-major by default
      strides = np.array([img.shape[1], 1, img.shape[1], 1])
      if self.transpose:
         strides = np.array([img.shape[1], 1, 1, img.shape[1]])
      # the byte length of each item in img
      strides = strides*img.itemsize
      # form the four dimensional matrix
      blocks = np.lib.stride_tricks.as_strided(img, shape=shape, strides=strides)
      if keepPos:
           return blocks.reshape(shape[0], shape[1], pSize*pSize)
      else:
           return blocks.reshape(shape[0]*shape[1], pSize*pSize).T

   @staticmethod
   def display(img,title,cmap="gray"):
      pylab.imshow(img, cmap)
      pylab.title(title)
      pylab.axis("off")
      pylab.show()

   @staticmethod
   def showAtoms(atoms,title,enhance=True,minCols=30,padding=1):
      dSize, nAtoms = atoms.shape
      pSize = int(np.sqrt(dSize))
      # number of atom rows and cols to be displayed
      cols = min([minCols, nAtoms])
      rows = int(nAtoms/cols)
      # enhance variation for visual appreciation
      if enhance:
         minA = atoms-np.tile(np.amin(atoms, axis=0), [dSize, 1])
         atoms = (minA/np.tile(np.amax(minA, axis=0), [dSize, 1]))*255
      # number of rows and cols of the image to be displayed
      nCols = cols*(pSize+padding)+padding
      nRows = rows*(pSize+padding)+padding
      # raster image has a white background
      raster = np.ones([nRows, nCols])*255
      # fill up the raster image
      for row in range(rows):
         for col in range(cols):
            raster[ padding+row*(padding+pSize):padding+row*(padding+pSize)+pSize,
               padding+col*(padding+pSize):padding+col*(padding+pSize)+pSize ] = atoms[:, row*cols+col].reshape(pSize, pSize).T
      full_title = "{} {}/{}".format(title, rows*cols, nAtoms)
      NN.display(raster, full_title)

   @staticmethod
   def showCDF(vals, what):
      xvals = np.sort(vals.reshape(1, len(vals)))
      yvals = np.cumsum(np.ones(xvals.shape))/len(vals)
      yvals = yvals.reshape(1,len(yvals))
      # remove duplicates
      ndups = [k for k in range(xvals.shape[1]-1) if (xvals[0, k] < xvals[0, k+1])] + [xvals.shape[1]-1]
      xvals = [xvals[0, k] for k in ndups]
      yvals = [yvals[0, k] for k in ndups]
      pylab.plot(xvals, yvals, "ro-")
      pylab.title("cumulative distribution function of " + what)
      pylab.show()

   @staticmethod
   def extendIm(img, margin):
      # extend the image horizontally
      lpart = np.fliplr(img[:, 1:margin+1])
      rpart = np.fliplr(img[:, -1-margin:-1])
      img = np.concatenate((lpart, img, rpart), axis=1)
      # extend the image vertically
      tpart = np.flipud(img[1:margin+1, :])
      bpart = np.flipud(img[-1-margin:-1, :])
      img = np.concatenate((tpart, img, bpart), axis=0)
      return img



if __name__ == "__main__":

   np.random.seed(100)
   net = "GPU_NN"
   sigma = 10
   nn = NN(net, sigma)
   nn.showNN()
   # nn.denoiseEmAll("IPOL", 15, freshRun=False)

   # single image case
   clean_folder = "test"
   name = "aface.png"
   clean = scipy.misc.imread(os.path.join(clean_folder, name)).astype(np.single)
   nn.neuralDenoising(clean, sigma, name)
